@@ -8,6 +8,12 @@
 /* Define this to aid in finding memory leaks */
 /*#define MAL_MEM_DEBUG*/
 /*#define MAL_DEBUG*/
+/*#define MAL_REF_DEBUG*/
+
+/* Logging file used by debugging facility */
+#ifndef MAL_DEBUG_OUTPUTFILE
+# define MAL_DEBUG_OUTPUTFILE "mxTools.log"
+#endif
 
 /* We want all our symbols to be exported */
 #define MX_BUILDING_MXTOOLS
@@ -15,7 +21,20 @@
 #include "mx.h"
 #include "mxTools.h"
 
-#define VERSION "0.8"
+/* Needed for mxTools_EvalCodeString() */
+#include "compile.h"
+#include "eval.h"
+
+/* Needed for mxTools_cur_frame() */
+#include "frameobject.h"
+
+#define VERSION "1.0.0"
+
+/* To open up the playground for new ideas... */
+/*#define INCLUDE_FUNSTUFF*/
+
+/* Maximal size of the code cache used by mxTools_EvalCodeString() */
+#define MAX_CODECACHE_SIZE	10
 
 /* --- module doc-string -------------------------------------------------- */
 
@@ -23,7 +42,7 @@ static char *Module_docstring =
 
  MXTOOLS_MODULE" -- A tool collection. Version "VERSION"\n\n"
 
- "(c) Copyright Marc-Andre Lemburg, 1997, mal@lemburg.com,\n\n"
+ "(c) Copyright Marc-Andre Lemburg, 1997,1998, mal@lemburg.com,\n\n"
  "                 All Rights Reserved\n\n"
  "See the documentation for further information on copyrights,\n"
  "or contact the author."
@@ -34,6 +53,12 @@ static char *Module_docstring =
 static PyObject *mxTools_Error;			/* Error Exception
 						   object */
 
+static PyObject *mxNotGiven;			/* NotGiven singleton */
+
+static PyObject *mxTools_BaseobjAttribute;      /* Attribute to use for
+						   acquire() per default */
+
+
 /* --- forward declarations ----------------------------------------------- */
 
 /* --- internal macros ---------------------------------------------------- */
@@ -72,6 +97,19 @@ void insint(PyObject *dict,
     PyDict_SetItemString(dict, name, v);
     Py_XDECREF(v);
 }
+#endif
+
+/* Helper for adding string constants to a dictionary. Check for
+   errors with PyErr_Occurred() */
+static 
+void insstr(PyObject *dict,
+	    char *name,
+	    char *value)
+{
+    PyObject *v = PyString_FromString(value);
+    PyDict_SetItemString(dict, name, v);
+    Py_XDECREF(v);
+}
 
 static
 PyObject *notimplemented1(PyObject *v)
@@ -99,12 +137,58 @@ PyObject *notimplemented3(PyObject *u, PyObject *v, PyObject *w)
  onError:
     return NULL;
 }
+
+/* --- Internal Functions ------------------------------------------------- */
+
+#ifdef INCLUDE_FUNSTUFF
+static
+PyObject *mxTools_EvalCodeString(PyObject *codestr)
+{
+    static PyObject *codecache;
+    PyObject *code;
+    PyObject *v;
+    
+    Py_Assert(PyString_Check(codestr),
+	      PyExc_SystemError,
+	      "Bad internal call to mxTools_EvalCodeString");
+
+    /* Init. codecache dictionary */
+    if (codecache == NULL) {
+	codecache = PyDict_New();
+	if (!codecache)
+	    goto onError;
+    }
+    
+    /* Get code object or compile the string */
+    code = PyDict_GetItem(codecache,codestr);
+    if (!code) {
+	code = Py_CompileString(PyString_AS_STRING(codestr),
+				"<string>", Py_eval_input);
+	if (!code)
+	    goto onError;
+	if (PyDict_Size(codecache) >= MAX_CODECACHE_SIZE)
+	    PyDict_Clear(codecache);
+	PyDict_SetItem(codecache,codestr,code);
+    }
+    else
+	Py_INCREF(code);
+    
+    /* Run the code in the current context */
+    v = PyEval_EvalCode((PyCodeObject *)code,
+			PyEval_GetGlobals(),
+			PyEval_GetLocals());
+    Py_DECREF(code);
+    return v;
+
+ onError:
+    return NULL;
+}
 #endif
 
-/* --- Tools --------------------------------------------------------*/
+/* --- Interface ---------------------------------------------------------- */
 
 Py_C_Function( mxTools_napply,
-	       "napply(number_of_calls,function,args=(),kw={})\n"
+	       "napply(number_of_calls,function,args=(),kw={})\n\n"
 	       "Calls the function number_of_calls times with the same\n"
 	       "arguments and returns a tuple with the return values.")
 {
@@ -188,7 +272,7 @@ Py_C_Function( mxTools_napply,
 }
 
 Py_C_Function( mxTools_mapply,
-	       "mapply(callable_objects,args=(),kw={})\n"
+	       "mapply(callable_objects,args=(),kw={})\n\n"
 	       "Calls the callable_objects in the given order with the same\n"
 	       "arguments and returns a tuple with the return values.")
 {
@@ -290,7 +374,7 @@ Py_C_Function( mxTools_mapply,
 }
 
 Py_C_Function( mxTools_method_mapply,
-	       "method_mapply(objects,methodname,args=(),kw={})\n"
+	       "method_mapply(objects,methodname,args=(),kw={})\n\n"
 	       "Calls the method methodname of all objects in the given\n"
 	       "order with the same arguments and returns a tuple with\n"
 	       "the return values.")
@@ -403,7 +487,7 @@ Py_C_Function( mxTools_method_mapply,
 }
 
 Py_C_Function( mxTools_trange,
-	       "trange([start=0,]stop[,step=1])\n"
+	       "trange([start=0,]stop[,step=1])\n\n"
 	       "Returns tuple(range(start,stop,step))")
 {
     int start,stop=INT_MAX,step=INT_MAX;
@@ -472,7 +556,7 @@ Py_C_Function( mxTools_trange,
 }
 
 Py_C_Function( mxTools_indices,
-	       "indices(object)\n"
+	       "indices(object)\n\n"
 	       "Returns tuple(range(len(object))).")
 {
     register int n;
@@ -539,7 +623,7 @@ Py_C_Function( mxTools_range_len,
 }
 
 Py_C_Function( mxTools_irange,
-	       "irange(object[,indices])\n"
+	       "irange(object[,indices])\n\n"
 	       "Returns a tuple of tuples (index,object[index]), one\n"
 	       "for each item in the sequence indices or, if this is not\n"
 	       "given, in trange(len(object)).")
@@ -624,17 +708,17 @@ Py_C_Function( mxTools_irange,
 }
 
 Py_C_Function( mxTools_get,
-	       "get(object,index[,default])\n"
+	       "get(object,index[,default])\n\n"
 	       "Returns object[index], or, if that fails, default.")
 {
     PyObject *x, *w, *index;
-    PyObject *def = 0;
+    PyObject *def = mxNotGiven;
 
     Py_Get3Args("OO|O",w,index,def);
 
     x = PyObject_GetItem(w,index);
     if (!x) {
-	if (def) {
+	if (def != mxNotGiven) {
 	    PyErr_Clear();
 	    x = def;
 	    Py_INCREF(x);
@@ -648,14 +732,14 @@ Py_C_Function( mxTools_get,
     return NULL;
 }
 
-Py_C_Function( mxTools_mget,
-	       "mget(object,indices[,defaults])\n"
+Py_C_Function( mxTools_extract,
+	       "extract(object,indices[,defaults])\n\n"
 	       "Returns a list of entries object[index] for each index\n"
 	       "in the sequence indices. defaults must have the same length\n"
 	       "as indices and is used to provide default values in case\n"
 	       "the lookup fails.")
 {
-    register int n = 0;
+    int n = 0;
     register int index;
     PyObject *t = 0;
     PyObject *w;
@@ -686,12 +770,12 @@ Py_C_Function( mxTools_mget,
 		goto onError;
 	    }
 	    x = PyObject_GetItem(w,v);
+	    Py_DECREF(v);
 	    if (!x) {
 		/* Use default value */
 		PyErr_Clear();
 		x = PySequence_GetItem(defaults,index);
 		if (!x) {
-		    Py_DECREF(v);
 		    PyErr_Format(PyExc_IndexError,
 			      "default value for index nr. %i not accessible",
 				 index);
@@ -713,8 +797,8 @@ Py_C_Function( mxTools_mget,
 		goto onError;
 	    }
 	    x = PyObject_GetItem(w,v);
+	    Py_DECREF(v);
 	    if (!x) {
-		Py_DECREF(v);
 		PyErr_Format(PyExc_IndexError,
 			     "default value for index nr. %i not accessible",
 			     index);
@@ -730,8 +814,83 @@ Py_C_Function( mxTools_mget,
     return NULL;
 }
 
+Py_C_Function( mxTools_iremove,
+	       "iremove(object,indices)\n\n"
+	       "Removes the items indexed by indices from object.\n"
+	       "For sequences the index list must be sorted ascending;\n"
+	       "an IndexError will be raised otherwise (object is then\n"
+	       "left in an undefined state)."
+	       )
+{
+    int n = 0;
+    register int index;
+    PyObject *w;
+    PyObject *indices;
+
+    Py_Get2Args("OO",w,indices);
+
+    n = PyObject_Length(indices);
+    if (n < 0)
+	Py_Error(PyExc_TypeError,
+		 "second argument must be a sequence");
+    
+    if (PyMapping_Check(w))
+	for (index = n - 1; index >= 0; index--) {
+	    register PyObject *v;
+	    register int rc;
+	    
+	    v = PySequence_GetItem(indices,index);
+	    if (!v) {
+		PyErr_Format(PyExc_IndexError,
+			     "index %i not accessible",
+			     index);
+		goto onError;
+	    }
+	    rc = PyMapping_DelItem(w,v);
+	    Py_DECREF(v);
+	    if (rc)
+        	goto onError;
+	}
+
+    else if (PySequence_Check(w)) {
+	int prev_index = INT_MAX;
+	
+	for (index = n - 1; index >= 0; index--) {
+	    register PyObject *v;
+	    register int i;
+	    
+	    v = PySequence_GetItem(indices,index);
+	    if (!v && PyInt_Check(v)) {
+		PyErr_Format(PyExc_IndexError,
+			     "index %i not accessible or not an integer",
+			     index);
+		goto onError;
+	    }
+	    i = PyInt_AS_LONG(v);
+	    Py_DECREF(v);
+	    if (prev_index >= i) {
+		if (PySequence_DelItem(w,i))
+		    goto onError;
+		prev_index = i;
+	    }
+	    else
+		Py_Error(PyExc_IndexError,
+			 "indices must be sorted ascending for sequences");
+	}
+    }
+    
+    else
+	Py_Error(PyExc_TypeError,
+		 "object must be a mapping or a sequence");
+    
+    Py_ReturnNone();
+
+ onError:
+    return NULL;
+}
+
 Py_C_Function( mxTools_ifilter,
-	       "ifilter(condition,object[,indices])\n"
+	       "ifilter(condition,object[,indices])\n\n"
 	       "Returns a list of tuples (index,object[index]) such that\n"
 	       "condition(object[item]) is true and index is found in\n"
 	       "the sequence indices (defaulting to indices(object)).\n"
@@ -859,10 +1018,10 @@ Py_C_Function( mxTools_ifilter,
 }
 
 Py_C_Function( mxTools_tuples,
-	       "tuples(a,b,c,...)\n"
-	       "Returns a list much like map(None,a,b,c,...) does. Only\n"
-	       "with a bit more intuitive name. This function does not\n"
-	       "optimize for the case where the sequences are of different\n"
+	       "tuples(sequence)\n\n"
+	       "Returns a list much like apply(map,(None,)+tuple(sequence)))\n"
+	       "does. Only with a bit more intuitive name. This function does\n"
+	       "not optimize for the case where the sequences are of different\n"
 	       "size and the resulting list of tuples will always\n"
 	       "have the length of the first sequence. Missing entries\n"
 	       "from the other sequences are filled in with None."
@@ -870,51 +1029,74 @@ Py_C_Function( mxTools_tuples,
 {
     int n,m;
     register int i;
+    register int j;
     PyObject *l = 0;
-    PyObject *w;
+    PyObject *arg,*w;
 
-    w = PyTuple_GetItem(args,0);
-    Py_Assert(w,
-	      PyExc_TypeError,
-	      "tuples() requires at least one argument");
+    Py_GetArgObject(arg);
 
+    /* Get tuple size (m) */
+    m = PySequence_Length(arg);
+    Py_Assert(m > 0,
+              PyExc_TypeError,
+              "sequence must have at least one element");
+
+    /* Get list size (n) */
+    w = PySequence_GetItem(arg,0);
+    if (!w)
+	goto onError;
     n = PySequence_Length(w);
-    if (n < 0)
-	Py_Error(PyExc_TypeError,
-		 "arguments must be sequences");
+    Py_DECREF(w);
+    Py_Assert(n >= 0,
+	      PyExc_TypeError,
+	      "sequence elements must be sequences");
 
+    /* XXX Could speed this up by rearranging and joining the loops */
+
+    /* Create list of tuples */
     l = PyList_New(n);
     if (!l)
 	goto onError;
-    
-    m = PyTuple_GET_SIZE(args);
+    for (j = 0; j < n; j++) {
+        PyObject *v;
+        
+        v = PyTuple_New(m);
+	if (!v)
+	    goto onError;
+	PyList_SET_ITEM(l,j,v);
+    }
 
-    for (i = 0; i < n; i++) {
-	    PyObject *t;
-	    register int j;
-
-	    t = PyTuple_New(m);
-	    if (!t)
-		goto onError;
+    /* Fill them in */
+    for (i = 0; i < m; i++) {
+	PyObject *u;
+	
+	u = PySequence_GetItem(arg,i);
+	if (!u)
+	    goto onError;
+	
+	for (j = 0; j < n; j++) {
+	    PyObject *v;
 	    
-	    for (j = 0; j < m; j++) {
-		register PyObject *v;
-		
-		v = PySequence_GetItem(PyTuple_GET_ITEM(args,j), i);
-		if (!v) {
-		    if (PyErr_ExceptionMatches(PyExc_IndexError)) {
-			PyErr_Clear();
-			v = Py_None;
-			Py_INCREF(v);
+	    v = PySequence_GetItem(u,j);
+	    if (!v) {
+		if (PyErr_ExceptionMatches(PyExc_IndexError)) {
+		    PyErr_Clear();
+		    /* Fill up the rest with None */
+		    for (; j < n; j++) {
+			Py_INCREF(Py_None);
+			PyTuple_SET_ITEM(PyList_GET_ITEM(l,j),i,Py_None);
 		    }
-		    else
-			goto onError;
+		    break;
+		}
+		else {
+		    Py_DECREF(u);
+		    goto onError;
 		}
-		PyTuple_SET_ITEM(t,j,v);
 	    }
-	    PyList_SET_ITEM(l,i,t);
+	    PyTuple_SET_ITEM(PyList_GET_ITEM(l,j),i,v);
+	}
+	Py_DECREF(u);
     }
-    
     return l;
 
  onError:
@@ -923,72 +1105,80 @@ Py_C_Function( mxTools_tuples,
 }
 
 Py_C_Function( mxTools_lists,
-	       "lists(seq)\n"
-	       "Inverse of tuples(), except that None entries are not\n"
-	       "interpreted as being placeholders for missing entries.\n"
-	       "seq must be a sequence of sequences, e.g. a\n"
-	       "list of tuples. These are then converted into a tuple of\n"
-	       "lists. As with tuples(), the first item in the sequence\n"
-	       "decides how many lists will be created. All other items\n"
-	       "must have at least as many entries."
+	       "lists(sequence)\n\n"
+	       "Same as tuples(), except that a tuple of lists is created."
 	       )
 {
     int n,m;
     register int i;
+    register int j;
     PyObject *t = 0;
-    PyObject *w,*v;
+    PyObject *arg,*w;
 
-    Py_GetArgObject(w);
+    Py_GetArgObject(arg);
 
-    /* n is the length of all lists */
-    n = PySequence_Length(w);
+    /* Get list size (n) */
+    n = PySequence_Length(arg);
     Py_Assert(n > 0,
-	      PyExc_TypeError,
-	      "argument must be a sequence with at least one item");
+              PyExc_TypeError,
+              "sequence must have at least one element");
 
-    /* m is the number of lists in the tuple */
-    v = PySequence_GetItem(w,0);
-    if (!v)
-	goto onError;
-    m = PySequence_Length(v);
-    Py_DECREF(v);
-    if (m < 0)
+    /* Get tuple size (m) */
+    w = PySequence_GetItem(arg,0);
+    if (!w)
 	goto onError;
+    m = PySequence_Length(w);
+    Py_DECREF(w);
+    Py_Assert(m >= 0,
+	      PyExc_TypeError,
+	      "sequence elements must be sequences");
+
+    /* XXX Could speed this up by rearranging and joining the loops */
     
-    /* Create lists */
+    /* Create tuple of lists */
     t = PyTuple_New(m);
     if (!t)
-	goto onError;
-    for (i = 0; i < m; i++) {
-	register PyObject *v;
-	
-	v = PyList_New(n);
-	if (!v)
-	    goto onError;
-	PyTuple_SET_ITEM(t,i,v);
+        goto onError;
+    for (j = 0; j < m; j++) {
+        PyObject *v;
+        
+        v = PyList_New(n);
+        if (!v)
+            goto onError;
+        PyTuple_SET_ITEM(t,j,v);
     }
 
     /* Fill them in */
     for (i = 0; i < n; i++) {
-	    register int j;
-	    register PyObject *u;
-	    
-	    u = PySequence_GetItem(w,i);
-	    if (!u)
-		goto onError;
-	    for (j = 0; j < m; j++) {
-		register PyObject *v;
-		
-		v = PySequence_GetItem(u,j);
-		if (!v) {
+	PyObject *u;
+            
+	u = PySequence_GetItem(arg,i);
+	if (!u)
+	    goto onError;
+
+	for (j = 0; j < m; j++) {
+	    PyObject *v;
+                
+	    v = PySequence_GetItem(u,j);
+	    if (!v) {
+		if (PyErr_ExceptionMatches(PyExc_IndexError)) {
+		    PyErr_Clear();
+		    /* Fill up the rest with None */
+		    for (; j < n; j++) {
+			Py_INCREF(Py_None);
+			PyList_SET_ITEM(PyTuple_GET_ITEM(t,j),i,Py_None);
+		    }
+		    break;
+		}
+		else {
 		    Py_DECREF(u);
 		    goto onError;
 		}
-		PyList_SET_ITEM(PyTuple_GET_ITEM(t,j),i,v);
 	    }
-	    Py_DECREF(u);
+	    PyList_SET_ITEM(PyTuple_GET_ITEM(t,j),i,v);
+	}
+	Py_DECREF(u);
     }
-    
     return t;
 
  onError:
@@ -997,7 +1187,7 @@ Py_C_Function( mxTools_lists,
 }
 
 Py_C_Function( mxTools_count,
-	       "count(condition,sequence)\n"
+	       "count(condition,sequence)\n\n"
 	       "Count the number of objects in sequence for which the\n"
 	       "selection function condition returns true and return the\n"
 	       "result as integer.")
@@ -1104,7 +1294,7 @@ Py_C_Function( mxTools_exists,
 }
 
 Py_C_Function( mxTools_forall,
-	       "forall(condition,sequence)\n"
+	       "forall(condition,sequence)\n\n"
 	       "Return 1 if and only if condition is true for all\n"
 	       "of the items in sequence and 0 otherwise. condition\n"
 	       "must be a callable object.")
@@ -1159,7 +1349,7 @@ Py_C_Function( mxTools_forall,
 }
 
 Py_C_Function( mxTools_index,
-               "index(condition,sequence)\n"
+               "index(condition,sequence)\n\n"
                "Return the index of the first item for which condition\n"
 	       "returns true. A ValueError is raised in case no item\n"
 	       "is found.")
@@ -1217,7 +1407,7 @@ Py_C_Function( mxTools_index,
 }
 
 Py_C_Function( mxTools_sizeof,
-	       "sizeof(object)\n"
+	       "sizeof(object)\n\n"
 	       "Returns the size in memory of the object in bytes.\n"
 	       "Note that this doesn't show any extra space allocated by\n"
 	       "the object.")
@@ -1238,7 +1428,7 @@ Py_C_Function( mxTools_sizeof,
 }
 
 Py_C_Function( mxTools_findattr,
-	       "findattr(objectlist,attrname)\n"
+	       "findattr(objectlist,attrname)\n\n"
 	       "Returns the first attribute with name attrname found\n"
 	       "among the objects in the list.")
 {
@@ -1259,20 +1449,20 @@ Py_C_Function( mxTools_findattr,
 	      "second argument must be a string");
 
     for(i = 0; i < length; i++) {
-	register PyObject *v;
-	register PyObject *w;
+	PyObject *v;
+	PyObject *w;
 
 	v = PySequence_GetItem(list,i);
 	if (!v)
 	    goto onError;
 
 	w = PyObject_GetAttr(v,name);
-	if (w) {
-	    PyErr_Clear();
+	if (w)
 	    return w;
-	}
 	else if (!PyErr_ExceptionMatches(PyExc_AttributeError))
 	    goto onError;
+	else
+	    PyErr_Clear();
     }
     Py_Error(PyExc_AttributeError,
 	     PyString_AS_STRING(name));
@@ -1280,8 +1470,59 @@ Py_C_Function( mxTools_findattr,
     return NULL;
 }
 
+Py_C_Function( mxTools_attrlist,
+	       "attrlist(objectlist,attrname)\n\n"
+	       "Returns a list of all attributes with the given name\n"
+	       "found among the objects in objectlist."
+	       )
+{
+    PyObject *list,*l=0;
+    PyObject *name;
+    register int i;
+    int length;
+
+    Py_Get2Args("OO",list,name);
+
+    length = PySequence_Length(list);
+    if (length < 0)
+	Py_Error(PyExc_TypeError,
+		 "first argument must be a sequence");
+
+    Py_Assert(PyString_Check(name),
+	      PyExc_TypeError,
+	      "second argument must be a string");
+
+    l = PyList_New(0);
+    if (!l)
+	goto onError;
+
+    for(i = 0; i < length; i++) {
+	PyObject *v;
+	PyObject *w;
+
+	v = PySequence_GetItem(list,i);
+	if (!v)
+	    goto onError;
+
+	w = PyObject_GetAttr(v,name);
+	if (w) {
+	    PyList_Append(l,w);
+	    Py_DECREF(w);
+	}
+	else if (!PyErr_ExceptionMatches(PyExc_AttributeError))
+	    goto onError;
+	else
+	    PyErr_Clear();
+    }
+    return l;
+
+ onError:
+    Py_XDECREF(l);
+    return NULL;
+}
+
 Py_C_Function( mxTools_dict,
-	       "dict(seq)\n"
+	       "dict(seq)\n\n"
 	       "Creates a dictionary from the given items sequence.\n"
 	       "The sequence must contain sub-sequences of at least length 2,\n"
 	       "the first entry being interpreted as the key and the second as\n"
@@ -1332,8 +1573,56 @@ Py_C_Function( mxTools_dict,
     return NULL;
 }
 
+Py_C_Function( mxTools_setdict,
+	       "setdict(seq,value=None)\n\n"
+	       "Creates a dictionary from the given items sequence.\n"
+	       "The sequence must hashable entries which are used as\n"
+	       "dictionary keys. The values are all set to value."
+	       )
+{
+    int n;
+    PyObject *seq,*d = 0,*value = 0;
+    register int i;
+
+    Py_Get2Args("O|O",seq,value);
+    n = PySequence_Length(seq);
+    if (n < 0)
+	Py_Error(PyExc_TypeError,
+		 "argument must be a sequence");
+
+    if (!value) {
+	Py_INCREF(Py_None);
+	value = Py_None;
+    }
+    else
+	Py_INCREF(value);
+
+    d = PyDict_New();
+    if (!d)
+	goto onError;
+
+    for (i = 0; i < n; i++) {
+	register PyObject *k;
+
+	k = PySequence_GetItem(seq,i);
+	if (!k)
+	    goto onError;
+	if (PyDict_SetItem(d,k,value))
+	    goto onError;
+	Py_DECREF(k);
+    }
+
+    Py_DECREF(value);
+    return d;
+
+ onError:
+    Py_XDECREF(d);
+    Py_XDECREF(value);
+    return NULL;
+}
+
 Py_C_Function( mxTools_invdict,
-	       "invdict(d)\n"
+	       "invdict(d)\n\n"
 	       "Creates a dictionary with inverse mappings from the\n"
 	       "given dictionary d.")
 {
@@ -1363,7 +1652,7 @@ Py_C_Function( mxTools_invdict,
 }
 
 Py_C_Function( mxTools_reverse,
-	       "reverse(seq)\n"
+	       "reverse(seq)\n\n"
 	       "Creates a new sequence with reversed order of the items in\n"
 	       "seq. If seq is a tuple, then a tuple is returned. Otherwise\n"
 	       "a list is returned.")
@@ -1437,35 +1726,288 @@ Py_C_Function( mxTools_reverse,
     return NULL;
 }
 
-#if 0
-Py_C_Function( mxTools_preallocate,
-	       "preallocate(x)\n")
+#ifdef INCLUDE_FUNSTUFF
+static
+void mxTools_free(void *p)
+{
+    free(p);
+}
+
+Py_C_Function( mxTools_malloc,
+	       "malloc(x)\n\n")
 {
     int x;
-    char *m;
+    char *p;
     register int i;
     register char *c;
 
     Py_GetSingleArg("i",x);
+
     /* Allocate virtual memory */
-    m = (char *)malloc(x);
-    if (!m) {
+    p = (char *)malloc(x);
+    if (!p) {
 	PyErr_NoMemory();
 	goto onError;
     }
+
     /* Turn the virtual memory into "real" one */
     i = x;
-    c = m;
+    c = p;
     while (i--)
 	*c++ = '\0';
-    /* Now free that space again */
-    free(m);
-    Py_ReturnNone();
+
+    return PyCObject_FromVoidPtr(p,mxTools_free);
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxTools_caching_eval,
+	       "caching_eval(code_string)\n\n"
+	       )
+{
+    PyObject *codestr;
+
+    Py_GetArg("O",codestr);
+
+    Py_Assert(PyString_Check(codestr),
+	      PyExc_TypeError,
+	      "arg must be a string");
+
+    return mxTools_EvalCodeString(codestr);
+
  onError:
     return NULL;
 }
 #endif
 
+Py_C_Function( mxTools_acquire,
+	       "acquire(self,name[,baseobjattr='baseobj'])\n\n"
+	       "Tries to get the attribute name from self.<baseobjattr>.\n"
+	       "If this is not defined or None, an AttributeError is\n"
+	       "raised. Otherwise getattr(self.<baseobjattr>,name) is\n"
+	       "returned. Attribute names must not start with an\n"
+	       "underscore (this too raises an AttributeError).\n")
+{
+    PyObject *obj,*baseobj,*name;
+    PyObject *v,*baseobjattr = mxTools_BaseobjAttribute;
+
+    static int recdepth;
+
+    recdepth++;
+    Py_Assert(recdepth < 2000,
+	      PyExc_SystemError,
+	      "maximum acquire() recursion depth exceeded");
+
+    Py_Get3Args("OO|O",obj,name,baseobjattr);
+
+    Py_Assert(PyString_Check(name),
+	      PyExc_TypeError,
+	      "attribute name must be a string");
+    
+    /* We don't acquire names starting with underscores */
+    Py_Assert(PyString_AS_STRING(name)[0] != '_',
+	      PyExc_AttributeError,
+	      PyString_AS_STRING(name));
+
+    /* baseobj = obj.<baseobjattr> */
+    baseobj = PyObject_GetAttr(obj,baseobjattr);
+    if (!baseobj || baseobj == Py_None) {
+	Py_XDECREF(baseobj);
+	Py_Error(PyExc_AttributeError,
+		 PyString_AS_STRING(name));
+    }
+    
+    /* Now return getattr(baseobj,name) */
+    v = PyObject_GetAttr(baseobj,name);
+    Py_DECREF(baseobj);
+    recdepth--;
+    return v;
+
+ onError:
+    recdepth--;
+    return NULL;
+}
+
+Py_C_Function( mxTools_verbosity,
+	       "verbosity([level])\n\n"
+	       "Sets the value of the interpreter's verbosity flag.\n"
+	       "Returns the flag's value before changing it or, when called\n"
+	       "without level, the current value."
+	       )
+{
+    int value = Py_VerboseFlag,
+	old_value = value;
+
+    Py_GetArg("|i",value);
+    Py_VerboseFlag = value;
+
+    return PyInt_FromLong((long)old_value);
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxTools_optimization,
+	       "optimization([level])\n\n"
+	       "Sets the value of the interpreter's optimization flag.\n"
+	       "Returns the flag's value before changing it or, when called\n"
+	       "without level, the current value."
+	       )
+{
+    int value = Py_OptimizeFlag,
+	old_value = value;
+
+    Py_GetArg("|i",value);
+    Py_OptimizeFlag = value;
+
+    return PyInt_FromLong((long)old_value);
+
+ onError:
+    return NULL;
+}
+
+Py_C_Function( mxTools_cur_frame,
+	       "cur_frame([offset=0])\n\n"
+	       "Returns the current execution frame, optionally going up the\n"
+	       "stack by offset levels. If there are less than offset\n"
+	       "frames on the stack, None is returned. The function is thread\n"
+	       "safe."
+	       )
+{
+    PyFrameObject *frame = NULL;
+    PyObject *v;
+    int offset = 0;
+
+    Py_GetArg("|i",offset);
+
+    /* Get requested frame object */
+    frame = (PyFrameObject *)PyEval_GetFrame();
+    for (;offset > 0;offset--) {
+	if (frame == NULL)
+	    break;
+	frame = frame->f_back;
+    }
+
+    /* Return frame or None */
+    v = (PyObject *)frame;
+    if (v == NULL)
+	v = Py_None;
+    Py_INCREF(v);
+    return v;
+
+ onError:
+    return NULL;
+}
+
+#if 0
+
+/* XXX replace with a generic protocol query API, e.g. 
+
+   hasmethods(obj,('__getitem__','__len__'))
+
+   which works for both instances *and* types !
+
+*/
+
+Py_C_Function( mxTools_issequence,
+	       "issequence(obj)\n\n"
+	       "Returns 1 iff obj exposes the sequence slot protocol,\n"
+	       "and at least defines the __getitem__ method, 0 otherwise."
+	       )
+{
+    int rc;
+    PyObject *v;
+
+    Py_GetArg("O",v);
+    if (PyInstance_Check(v)) {
+	PyObject *w;
+	
+	w = PyObject_GetAttrString("__getitem__");
+	if (!w) {
+	    PyErr_Clear();
+	    rc = 0;
+	}
+	else {
+	    Py_DECREF(w);
+	    rc = 1;
+	}
+    }
+    else {
+	rc = (v->ob_type->tp_as_sequence &&
+	      v->ob_type->tp_as_sequence->sq_item);
+    }
+    return rc;
+
+ onError:
+    return NULL;
+}
+#endif
+
+/* --- NotGiven Singleton -------------------------------------------------- */
+
+static 
+PyObject *mxNotGiven_Repr(PyObject *v)
+{
+    return PyString_FromString("NotGiven");
+}
+
+static 
+int mxNotGiven_NonZero(PyObject *v)
+{
+    return 0;
+}
+
+typedef PyObject mxNotGivenObject;
+
+static
+PyNumberMethods mxNotGiven_TypeAsNumber = {
+
+    /* These slots are not NULL-checked, so we must provide dummy functions */
+    notimplemented2,				/*nb_add*/
+    notimplemented2,				/*nb_subtract*/
+    notimplemented2,				/*nb_multiply*/
+    notimplemented2,				/*nb_divide*/
+    notimplemented2,				/*nb_remainder*/
+    notimplemented2,				/*nb_divmod*/
+    notimplemented3,				/*nb_power*/
+    notimplemented1,				/*nb_negative*/
+    notimplemented1,				/*nb_positive*/
+
+    /* Everything below this line EXCEPT nb_nonzero (!) is NULL checked */
+    (unaryfunc)0,				/*nb_absolute*/
+    mxNotGiven_NonZero,				/*nb_nonzero*/
+    (unaryfunc)0,				/*nb_invert*/
+    (binaryfunc)0,				/*nb_lshift*/
+    (binaryfunc)0,				/*nb_rshift*/
+    (binaryfunc)0,				/*nb_and*/
+    (binaryfunc)0,				/*nb_xor*/
+    (binaryfunc)0,				/*nb_or*/
+    (coercion)0,				/*nb_coerce*/
+    (unaryfunc)0,				/*nb_int*/
+    (unaryfunc)0,				/*nb_long*/
+    (unaryfunc)0,				/*nb_float*/
+    (unaryfunc)0,				/*nb_oct*/
+    (unaryfunc)0,				/*nb_hex*/
+};
+
+static PyTypeObject mxNotGiven_Type = {
+    PyObject_HEAD_INIT(0)		/* init at startup ! */
+    0,		/*ob_size*/
+    "NotGiven",	/*tp_name*/
+    sizeof(mxNotGivenObject),		/*tp_basicsize*/
+    0,		/*tp_itemsize*/
+    0,		/*tp_dealloc; object always stays alive */
+    0,		/*tp_print*/
+    0,		/*tp_getattr*/
+    0,		/*tp_setattr*/
+    0,		/*tp_compare*/
+    mxNotGiven_Repr, /*tp_repr*/
+    &mxNotGiven_TypeAsNumber,		/*tp_as_number*/
+    0,		/*tp_as_sequence*/
+    0,		/*tp_as_mapping*/
+    0,		/*tp_hash */
+};
+
 /* --- module init --------------------------------------------------------- */
 
 /* Python Method Table */
@@ -1480,9 +2022,12 @@ PyMethodDef Module_methods[] =
     Py_MethodListEntry("irange",mxTools_irange),
     Py_MethodListEntry("ifilter",mxTools_ifilter),
     Py_MethodListEntry("get",mxTools_get),
-    Py_MethodListEntry("mget",mxTools_mget),
+    Py_MethodListEntry("extract",mxTools_extract),
+    Py_MethodListEntry("iremove",mxTools_iremove),
+    Py_MethodListEntry("mget",mxTools_extract),
     Py_MethodListEntry("mgetattr",mxTools_findattr),
     Py_MethodListEntry("findattr",mxTools_findattr),
+    Py_MethodListEntry("attrlist",mxTools_attrlist),
     Py_MethodListEntry("mapply",mxTools_mapply),
     Py_MethodListEntry("method_mapply",mxTools_method_mapply),
     Py_MethodListEntry("count",mxTools_count),
@@ -1493,42 +2038,74 @@ PyMethodDef Module_methods[] =
     Py_MethodListEntrySingleArg("sizeof",mxTools_sizeof),
     Py_MethodListEntrySingleArg("dict",mxTools_dict),
     Py_MethodListEntrySingleArg("invdict",mxTools_invdict),
+    Py_MethodListEntry("setdict",mxTools_setdict),
     Py_MethodListEntrySingleArg("reverse",mxTools_reverse),
-    Py_MethodListEntry("tuples",mxTools_tuples),
+    Py_MethodListEntrySingleArg("tuples",mxTools_tuples),
     Py_MethodListEntrySingleArg("lists",mxTools_lists),
-#if 0
-    Py_MethodListEntrySingleArg("preallocate",mxTools_preallocate),
+    Py_MethodListEntry("acquire",mxTools_acquire),
+    Py_MethodListEntry("verbosity",mxTools_verbosity),
+    Py_MethodListEntry("optimization",mxTools_optimization),
+    Py_MethodListEntry("cur_frame",mxTools_cur_frame),
+#ifdef INCLUDE_FUNSTUFF
+    Py_MethodListEntrySingleArg("malloc",mxTools_malloc),
+    Py_MethodListEntry("caching_eval",mxTools_caching_eval),
 #endif
     {NULL,NULL} /* end of list */
 };
 
+/* Cleanup function */
+static 
+void mxToolsModule_Cleanup(void)
+{
+    Py_XDECREF(mxTools_BaseobjAttribute);
+}
+
 /* create PyMethodObjects and register them in the module's dict */
 MX_EXPORT(void) 
      initmxTools(void)
 {
     PyObject *module, *moddict;
 
+    /* Init type objects */
+    PyType_Init(mxNotGiven_Type);
+
     /* create module */
     module = Py_InitModule4(MXTOOLS_MODULE, /* Module name */
 			    Module_methods, /* Method list */
 			    Module_docstring, /* Module doc-string */
 			    (PyObject *)NULL, /* always pass this as *self */
 			    PYTHON_API_VERSION); /* API Version */
+    if (!module)
+	goto onError;
+
+    /* Register cleanup function */
+    if (Py_AtExit(mxToolsModule_Cleanup))
+	/* XXX what to do if we can't register that function ??? */;
+
+    /* Create NotGiven singleton */
+    mxNotGiven = PyObject_NEW(PyObject,&mxNotGiven_Type);
+    if (!mxNotGiven)
+	goto onError;
+
+    /* Create acquire default argument for baseobjattr */
+    mxTools_BaseobjAttribute = PyString_InternFromString("baseobj");
+    if (!mxTools_BaseobjAttribute)
+	goto onError;
 
     /* Add some constants to the module's dict */
     moddict = PyModule_GetDict(module);
-    PyDict_SetItemString(moddict, 
-			 "__version__",
-			 PyString_FromString(VERSION));
+    insstr(moddict,"__version__",VERSION);
+    PyDict_SetItemString(moddict,
+			 "NotGiven",
+			 mxNotGiven);
 
     /* Errors */
     if (!(mxTools_Error = insexc(moddict,"Error")))
 	goto onError;
 
-    /* Check for errors */
+ onError:
+    /* Check for errors and report them */
     if (PyErr_Occurred())
-	Py_Error(PyExc_ImportError,
-		 "initialization of module "MXTOOLS_MODULE" failed");
-onError:
+	Py_ReportModuleInitError(MXTOOLS_MODULE);
     return;
 }
